class BooleanFunction

	attr_accessor :inputs, :state, :index, :name, :mode,
		:function, :failpercentage
	attr_reader :numinputs, :graphnode

	@@id = 0

	def initialize numinputs, initvalue, function = nil
		@inputs = Array.new(numinputs, nil)
		@numinputs = numinputs

		@state = initvalue
        @next_state = initvalue

        @function = function

		@name = nil

		@index = nil
		@@id += 1
		@id = "n" + @@id.to_s


        @mode = nil
	end

	def initialize_copy a
		super
		@inputs = Array.new(a.inputs.size, nil)
		@numinputs = a.numinputs
		@name = a.name
		@id = a.id
		@state = a.state
		@graphnode = a.graphnode
	end

	def id
		return @name if @name
		@id
	end

    def to_s
        "#{id}: #{@mode.to_s}"
    end

    def write_inputs graph
        @inputs.each do |input|
            graph.add_edge(@graphnode, input.graphnode)
        end
    end

	def addInput input
		@inputs.size.times do |i|
			if @inputs[i].nil?
				@inputs[i] = input
				return
			end
		end
		puts "Error, could not find free input for " + id()
	end

    def write_node graph
        @graphnode = graph.add_node(id())
    end

    def calculate_state
        if @function
            @next_state = @function.call(@inputs.map{|input| input.state})
        else
            @next_state = 0
        end
    end

    def commit_state
        case @mode
        when :On
            #always on state
            @state = 1
        when :Off
            #always off state
            @state = 0
        when :Random
            prob = @failpercentage || 1.0
            if rand < prob
                @state = (rand > 0.5)?0:1
            else
                @state = @next_state
            end
        when :Flip
            prob = @failpercentage || 1.0
            if rand < prob
                @state = case @next_state
                             when 0 then 1
                             when 1 then 0
                             end
            else
                @state = @next_state
            end

        else
            #in this case the node is working
            @state = @next_state
			if @state != 0 and @state != 1
				raise "State should be 0 or 1"
			end
        end
    end

end
