require "boolfunction.rb"
require 'rubygems'
require 'graphviz'

class BooleanNetwork

	attr_reader :nodes
	attr_accessor :description, :at_iterate

	def initialize
		@nodes = Array.new
		@index = 0
		@description = "Stock Boolean Network"
        @at_iterate = nil
	end

	def addNode node
		throw Exception.new("Wrong input") if not node.is_a?(BooleanFunction)
		node.index = @index
		@index += 1
		@nodes.push node
		return node
	end

	def clone
		n = self.class.new
		clones = {}
		
		#create clones
		@nodes.each do |node|
			clones[node] = n.addNode(node.clone)
		end

		#and restore the inputs
		nodes.each_with_index do |oldnode, node_index|
			oldnode.inputs.each_with_index do |oldinput, input_index|
				n.nodes[node_index].inputs[input_index] = 
					clones[oldinput]
			end
		end

		return n
	end

    def find id
        @nodes.find_all{|node| node.id == id}
    end

	def set_random_state
        @nodes.each do |node|
            node.state = (rand > 0.5)?0:1
        end
	end

	def set_state array
		a = array.clone
        @nodes.each do |node|
            node.state = a.shift
        end
	end

	def iterate method = :Synchronous

		case method
		when :Synchronous
			@nodes.each do |node|
				node.calculate_state
			end
			@nodes.each do |node|
				node.commit_state
			end
		when :Asynchronous
            rand_node = @nodes[@nodes.size*rand]
			rand_node.calculate_state
			rand_node.commit_state
		end

        if not @at_iterate.nil?
            @at_iterate.call(self)
        end
	end

	def stabilize stablethreshold, maxiter
		prevstate = statemap
		stable_times = 0

		maxiter.times do |iter|
			iterate :Asynchronous
            #printState
			state = statemap
			if state == prevstate
				stable_times += 1
			else
				stable_times = 1
				prevstate = state
			end
			if stable_times > stablethreshold
				return iter
			end
		end

        return maxiter
	end

	def setState stateArray
		@nodes.each do |node|
			node.state = stateArray.shift
		end
	end

    def statemap
        @nodes.map{|n| n.state}
    end

	def printState
		print "network state: ["
		@nodes.each do |node|
			if node.state != 0
				print "X"
			else
				print " "
			end
		end
		print "]\n"
	end

	def makegraph
        g = GraphViz.new( :G, :type => :digraph )
		#f.puts "digraph finite_state_machine {"
		#f.puts "	rankdir=LR;"
		#f.puts "	size=\"20,20\";"
		#f.puts "node [shape=circle];"
		@nodes.each do |node|
            node.write_node(g)
		end

		@nodes.each do |node|
            node.write_inputs(g)
		end
		return g
	end

	def writeGraph filename
        makegraph.output( :dot => filename )
	end

	def show
		filename = "tmp#{rand}.gv"
		makegraph.output( :dot => filename )
		if fork.nil? then
			exec("dot -Txlib #{filename}; rm #{filename}")
		end
	end

end
