#!/usr/bin/ruby

require "boolnetwork.rb"
require 'yeastnetwork.rb'
require "tablefunction.rb"
require 'typenetwork.rb'
require 'util.rb'
require 'zlib'

require 'optparse'
require 'ostruct'
#require 'ruby-debug'


#defaults go here
$opts = OpenStruct.new
$opts.networksize = 15
$opts.K = 3
$opts.output = "output"
$opts.experiment = :FailingTests
$opts.count = 50
$opts.param = []

OptionParser.new do |opts|
    opts.banner = "Usage: evolve.rb [options]"

    opts.on("-v", "--[no-]verbose", "Run verbosely") do |v|
        $opts.verbose = true
    end

    opts.on("-o", "--output FILENAME", "Output stats to FILENAME") do |file|
        $opts.output = file
    end

    opts.on("-e", "--experiment EXPERIMENT", "Perform EXPERIMENT") do |experiment|
        $opts.experiment = experiment.gsub(/\b\w/){$&.upcase}.to_sym
    end

    opts.on("-p", "--param PARAM", "Add extra experiment PARAMeter") do |param|
        $opts.param << param.gsub(/\b\w/){$&.upcase}.to_sym
    end

    opts.on("-n", "--netsize NETSIZE", "Set networksize to NETSIZE") do |netsize|
        $opts.networksize = netsize.to_i
    end

    opts.on("-c", "--count TIMES", "Run the iterations int he experiment this many times") do |count|
        $opts.count = count.to_i
    end

    opts.on_tail("-h", "--help", "Show this message") do
        puts opts
        exit
    end

end.parse!

$stop = false

$parallel_tests = [

	#differ as little as possible
	lambda{ |prev_state, state|
		add = 0.0
		prev_state.size.times do |i|
			if prev_state[i] == state[i]
				#add += 1
			end
		end
		add
	},
		
	#be on as much as possible
	lambda{ |prev_state, state|
		add = 0.0
		prev_state.size.times do |i|
			if state[i] == 0
				#add += 1
			end
		end
		add
	},
]

#if nil, do nothing. else, fill it
#set to empty array to start
$state_string = nil
$iterate_func = lambda{ |net|
    if not $state_string.nil?
        $state_string += net.statemap.join
    end
}

class BooleanNetwork

    def make_basins options = nil

        options ||= {}
        count = options[:BasinSize] || 200
        iteratecount = options[:IterateCount] || 200
        failtype = options[:FailType]
        failpercentage = options[:FailPercentage] || 0.5
        stabilizeperiod = options[:StabilizePeriod] || 0
        failednoderesetstate = options[:FailedNodeResetState] || 0

        basins = {}

		nodes.each { |somenode|
			somenode.failpercentage = failpercentage
		}

        count.times do |i|
            #print i.to_s + "\r"
            #$stdout.flush
            set_random_state

			nodes.each { |node| node.mode = failtype }

            iteratecount.times{ iterate(:Asynchronous) }

			if failtype and stabilizeperiod > 0
				nodes.each { |node| node.mode = nil }
				stabilizeperiod.times{ iterate(:Asynchronous) }
			end

            state = statemap
            basins[state] ||= 0
            basins[state] += 1
        end

		nodes.each { |somenode|
			somenode.mode = nil
		} if failtype

        return basins
    end

    def test_tolerance options = nil
        options ||= {}

        failcount = options[:FailCount] || 3
        basinsize = options[:BasinSize] || 200
        evalmethod = options[:EvaluationMethod] || :Average

        initbasins = make_basins(options)
        sortedinit = initbasins.sort{|a,b| b[1]<=>a[1]} if evalmethod == :CompareOrder

        #BooleanNetwork.print_basins(initbasins)
        sum = 0.0

        failcount.times do
            basins = make_basins(options)

            case evalmethod
            when :CompareOrder
                sorted = basins.sort{|a,b| b[1]<=>a[1]}
                #debugger
                sortedinit.each_with_index do |val, i|
                    break if not sorted[i]
                    if val[0] == sorted[i][0]
                        sum += val[1].to_f
                    end
                end
                sum /= basinsize
            when :Average
                score = BooleanNetwork.evaluate_basins(b)
                puts "failed basins: " + score.to_s
                sum += score
            end

        end
        puts sum / failcount
        return sum / failcount

    end

	def test_robustness
        basins = make_basins()
        return BooleanNetwork.evaluate_basins(basins)
    end
    
    def BooleanNetwork.print_basins basins
        basins.sort{|a,b| b[1]<=>a[1]}.each do |el|
            puts el[1].to_s.rjust(5) + ": " + el[0].map{|st|
                case st
                when 0 then " "
                when 1 then "X"
                end
            }.join
        end
    end

    def BooleanNetwork.evaluate_basins basins
        sum = 0.0
        basins.sort{|a,b| b[1]<=>a[1]}.each_with_index do |val, i|
            sum += val[1].to_f / (i + 1)
        end
        return sum
    end



	def write_basindist file
		add = 0.0

        endresult = make_basins(:BasinSize=>400)
        sorted_basins = endresult.sort{|a,b| b[1]<=>a[1]}

        #filter out irrelevant small basins
        while sorted_basins.last[1] < 5
            puts "removed " + sorted_basins.pop.join
        end

        sorted_basins.each do |el|
            puts el[1].to_s.rjust(5) + ": " + el[0].map{|st|
                case st
                when 0 then " "
                when 1 then "X"
                end
            }.join
        end

        return endresult
    end




    def test_complexity settings = nil
        @at_iterate = $iterate_func

		settings ||= {}
        compressed_size = 0
		original_size = 0

		arr = settings[:InitialState]
        mode = settings[:IterateMode] || :Asynchronous
        verbose = settings[:Verbose] || false

        (settings[:ResetCount] || 100).times do |i|
            print i.to_s + "\r"
            $stdout.flush

			if arr
				set_state(arr)
			else
				set_random_state
			end

            $state_string = ""

            (settings[:IterateCount] || 100).times{
				iterate(mode)
				printState if verbose
			}
            #stabilize(100, 1000)
            zlib = Zlib::Deflate.new()
            compressed_size += zlib.deflate($state_string, Zlib::FINISH).size
			original_size += $state_string.size
            zlib.close
            #puts $state_string
        end

        @at_iterate = nil

        return compressed_size.to_f / original_size
    end

   def test_intermediarylength settings = nil

		settings ||= {}

		arr = settings[:InitialState]
        threshold = settings[:Threshold] || 20
        maxiter = settings[:MaxIter] || 200
        resetcount = settings[:ResetCount] || 50

        sum = 0.0

        resetcount.times {
            set_random_state
            sum += stabilize(threshold, maxiter)
        }

        avg = sum / resetcount

        return avg
   end


	def evaluateNetwork
		numdiff = 0.0

        #numdiff += test_robustness
        #numdiff += test_complexity
        return -((64 - (test_complexity / 100)) ** 2 + (90 - test_intermediarylength) ** 2)

		return numdiff
	end

end

trap("INT") do
    if $stop
        exit 1
    else
        $stop = true
    end
end

stable = nil
largestdiff = -1.0/0

#File.open("log", "a") do |file|
    #file.puts Time.now.to_s + " " + $opts.to_s
#end

puts $opts

case $opts.experiment
when :Random
	#try different networks
	200.times do |neti|
		begin
			net = TypeNetwork.generate_network($opts.networksize)
		rescue RuntimeError => e
			puts e
			next
		end

		numdiff = net.evaluateNetwork
		puts "Network ##{neti}: #{numdiff}"

		if numdiff > largestdiff
			largestdiff = numdiff
			stable = net
		end

		break if $stop
	end

    puts "Largest: #{largestdiff}"
    stable.writeGraph("largest.gv")

when :Hillclimber
	stable = TypeNetwork.generate_network($opts.networksize)
    puts largestdiff

	#try different networks
	200.times do |neti|

		net = stable.clone
		net.randomize_nodes

		numdiff = net.test_complexity(
            :ResetCount=>20,
            :IterateCount=>50)
        puts "Network ##{neti}: #{numdiff}"

		if numdiff > largestdiff
            puts "=> (new high)"
			largestdiff = numdiff
			stable = net
		end

		break if $stop
	end

    puts "Largest: #{largestdiff}"
    stable.writeGraph("largest.gv")

when :QuadHillclimber
    output = File.new($opts.output, "a")
    loop do
        stable = TypeNetwork.generate_network($opts.networksize,
                                              :Types => [
                                                  :Or,
                                                  :And,
                                                  :Xor])
        largestscore = stable.evaluateNetwork

        #try different networks
        40.times do |neti|

            prevnet = stable.clone

            4.times do
                net = prevnet.clone
                net.randomize_nodes

                score = net.evaluateNetwork

                if score > largestscore
                    puts "Network ##{neti}: #{score}"
                    largestscore = score
                    stable = net
                end

                break if $stop
            end

            break if $stop
        end
        output.puts "#{stable.test_intermediarylength},  #{stable.test_complexity}"
        output.flush
    end


    #puts "Largest: #{largestscore}"
    #stable.writeGraph("largest.gv")



when :LocalityTest

    output = File.new($opts.output, "a")

	#try different networks
	loop do

        $distance_threshold = rand

		begin
			net = TypeNetwork.generate_network($opts.networksize)
		rescue RuntimeError => e
			puts e
			next
		end

		score = net.test_robustness
        puts score
        output.puts "#{score}, #{$distance_threshold}"
        output.flush

		break if $stop
	end
    output.close


when :RobustComplexity

    output = File.new($opts.output, "a")

	#try different networks
	loop do

		begin
			net = TypeNetwork.generate_network($opts.networksize)
		rescue RuntimeError => e
			puts e
			next
		end

		#score = net.evaluateNetwork
        #puts score
        output.puts "#{net.test_robustness},  #{net.test_complexity}"
        output.flush

		break if $stop
	end
    output.close

when :FailingTests
    output = File.new($opts.output, "a")

	#try different networks
	loop do

		begin
			net = TypeNetwork.generate_network($opts.networksize,
                                               :Types => [
                                                   :And,
                                                   :Or,
                                                   :Xor])
		rescue RuntimeError => e
			puts e
			next
		end

		#score = net.evaluateNetwork
        #puts score

        #flipprob = rand
        flipnode = net.nodes.random_item
        flipnode.mode = :Rand
        flipnode.failpercentage = rand

        output.puts "#{flipnode.failpercentage}, #{net.test_complexity}"
        output.flush

		break if $stop
	end
    output.close



when :InitialStates

	[   "111111111111111000000000000000",
		"111111110000000000000000000000",
		"111100000000000000000000000000",
		"100000000000000000000000000000",
		"101010101010101010101010101010",
		"000011110000111100001111000011"
		#"000000000000000000000000000000",
		#"111111111111111111111111111111",
	].each do |initstate|

		output = File.new($opts.output + "-#{initstate}", "a")
		puts "Using initstate " + initstate
		arr = []
		initstate.each_char{|char| arr.push char.to_i}

		netcount = 50

		#try different networks
		netcount.times do |neti|

			settings = {
				:ResetCount => 30,
				:IterateCount => 200,
				:InitialState => arr
			}

			begin
				net = TypeNetwork.generate_network($opts.networksize)
			rescue RuntimeError => e
				puts e
				next
			end

			#score = net.evaluateNetwork
			#puts score
			output.puts "#{net.test_robustness},  #{net.test_complexity(settings)}"
			output.flush
			puts "Finished net #{neti}/#{netcount}"

			break if $stop
		end
		output.close

		break if $stop
		
	end

when :Intermediary

    output = File.new($opts.output, "a")

	#try different networks
	loop do

		begin
			net = TypeNetwork.generate_network($opts.networksize,
                                               :Types => [
                                                   :And,
                                                   :Or,
                                                   :Xor])
		rescue RuntimeError => e
			puts e
			next
		end

		#score = net.evaluateNetwork
        #puts score
        interstatecount = net.test_intermediarylength(:Threshold=>10)
        output.puts "#{interstatecount},  #{net.test_complexity(
            :IterateCount=>interstatecount.to_i)}"
        output.flush

		break if $stop
	end
    output.close

when :BasinsDist

    output = File.new($opts.output, "a")

	#try different networks
	loop do

		begin
			net = TypeNetwork.generate_network($opts.networksize,
                                               :Types => [
                                                   :Threshold1,
                                                   :Threshold2,
                                                   :Threshold3])
		rescue RuntimeError => e
			puts e
			next
		end

        net.write_basindist(output)
        output.flush

		break if $stop
        break
	end
    output.close

when :NetworkSize

    output = File.new($opts.output, "a")

	#try different networks
	loop do

        size = 10 + (rand * 200.0).to_i
        puts size
		begin
			net = TypeNetwork.generate_network(size,
                                               :Types => [
                                                   :Threshold1,
                                                   :Threshold2,
                                                   :Threshold3])
		rescue RuntimeError => e
			puts e
			next
		end
        net.show
        break

        output.puts "#{size}, #{net.test_intermediarylength(:MaxIter => 800, :Threshold => 30)}"
        output.flush

		break if $stop
	end
    output.close
when :CAComplexity

    output = File.new($opts.output, "a")

    256.times do |rule|
        $carule = rule
        2.times do
            begin
                net = create_1d_ca($opts.networksize, lambda{ |states|
                    statepos = states[0] + (states[1] << 1) + (states[2] << 2)
                    if ($carule & (1<<statepos)) == 0
                        0
                    else
                        1
                    end
                })

            rescue RuntimeError => e
                puts e
                next
            end

            output.puts "#{rule}, #{net.test_complexity(:IterateMode => :Synchronous)}"
            output.flush

            break if $stop
        end
    end
    output.close
when :Tolerance
    $distance_threshold = 0.7
    net = TypeNetwork.generate_network($opts.networksize,
                                               :Types => [
                                                   :And,
                                                   :Or,
                                                   :Xor])
    #net.test_tolerance(
        #:BasinSize => 100)
        #:FailCount =>
    net.writeGraph("before.gv")

    5.times do 
        net = quad_ea(net, 5) { |newnet|
            newnet.test_robustness
        }
        net = quad_ea(net, 5) { |newnet|
            newnet.test_tolerance
        }
    end
    net.writeGraph("final.gv")
when :TolInter
    $distance_threshold = 0.7
    output = File.new($opts.output, "a")

    loop do 
        begin
            net = TypeNetwork.generate_network($opts.networksize,
                                               :Types => [
                                                   :And,
                                                   :Or,
                                                   :Xor])
        rescue RuntimeError => e
            puts e
            next
        end

        output.puts "#{net.test_intermediarylength()}, #{net.test_tolerance}"
        output.flush

        break if $stop
    end
    output.close
when :Boxplot

	$distance_threshold = 0.4
    output = File.new($opts.output, "a")

    combos = []

	#BN
    [ [:And, :Or, :Xor],
        [:Threshold1, :Threshold2, :Threshold3],
        [:NThreshold1, :NThreshold2, :NThreshold3]
    ].each do |types|

        [2,3,4].each do |k|

            [0.0].each do |failp|
                [false].each do |performea|

                    combos << Proc.new {
                        $opts.K = k
                        $opts.types = types
                        $opts.failpercentage = failp
                        $opts.performea = performea
						$opts.structure = :BooleanNetwork
                        "BN-#{($opts.performea)?"EA-":""}f#{failp.to_s}-K#{k.to_s}-#{types[0].to_s}"
                    }
                end
            end
        end
    end

	#CA
	
	dimension_size_type_combos = [
		[1, 64, [:Threshold1, :Threshold2, :Threshold3]],
		[1, 64, [:NThreshold1, :NThreshold2, :NThreshold3]],
		[2, 8, [:Threshold1, :Threshold2, :Threshold3, :Threshold4, :Threshold5]],
		[2, 8, [:NThreshold1, :NThreshold2, :NThreshold3, :NThreshold4, :NThreshold5]],
		[3, 4, [:Threshold1, :Threshold2, :Threshold3, :Threshold4, :Threshold5, :Threshold6, :Threshold7]],
		[3, 4, [:NThreshold1, :NThreshold2, :NThreshold3, :NThreshold4, :NThreshold5, :NThreshold6, :NThreshold7]],
	].each do |dstcombo|
		[0.0].each do |failp|
			[false].each do |performea|
				combos << Proc.new {

					$opts.dim = dstcombo[0]
					$opts.width = dstcombo[1]
					$opts.types = dstcombo[2]

					$opts.failpercentage = failp
					$opts.performea = performea
					$opts.structure = :CellularAutomata
					"CA-#{($opts.performea)?"EA-":""}f#{failp.to_s}-#{dstcombo[1]}x#{dstcombo[0]}D-#{dstcombo[2][0].to_s}"
				}
			end
		end
	end

    $opts.count.times do

        #get a random combination of params
        func = combos.random_item
        typeid = func.call
        puts $opts

        begin
			if $opts.structure == :BooleanNetwork
				net = TypeNetwork.generate_network($opts.networksize,
												   :Types => $opts.types,
												   :K => $opts.K)
			elsif $opts.structure == :CellularAutomata
				net = TypeNetwork.generate_ca($opts.width,
												   $opts.types,
												   $opts.dim)
			end
        rescue RuntimeError => e
            puts e
			next
        end

        tolerance_opts = { :FailType => :Random,
            :EvaluationMethod => :CompareOrder,
			:StabilizePeriod => 50,
			:IterateCount => 300,
            :FailPercentage => $opts.failpercentage }

        if $opts.performea
            net = quad_ea(net) do |newnet|
                newnet.test_tolerance(tolerance_opts)
            end
        end

		filterthreshold = 2

		if $opts.param.include?(:BasinCount)
			basins = net.make_basins(
				:IterateCount => 300,
				:BasinSize => 200
			)

			if $opts.param.include?(:FilterBasins)
				#filter out all basins that are too small
				basins = basins.find_all{ |b| b[1] > filterthreshold }
			end

			#BooleanNetwork.print_basins basins
			output.puts "#{typeid}, #{basins.size}"


		elsif $opts.param.include?(:FailBasins)
			5.times do
				failp = 0.4 * rand

				basinopts = {
					:IterateCount => 300,
					:BasinSize => 200,
					:FailType => :Random,
					:FailPercentage => failp
				}

				if $opts.param.include?(:StabilizePeriod)
					basinopts[:StabilizePeriod] = 50
				end

				failbasins = net.make_basins(basinopts)

				if $opts.param.include?(:FilterBasins)
					#filter out all basins that are too small
					failbasins = failbasins.find_all{ |b| b[1] > filterthreshold }
				end

				output.puts "#{typeid}, #{failp}, #{failbasins.size}"
			end

		elsif $opts.param.include?(:Intermediary)
			output.puts "#{typeid}, #{net.test_intermediarylength(:MaxIter=>400, :Threshold=>60)}"


		elsif $opts.param.include?(:Robustness)
			output.puts "#{typeid}, #{net.test_robustness}"


		elsif $opts.param.include?(:InterCompl)
			interstatecount = net.test_intermediarylength(:MaxIter=>300, :Threshold=>30)
			output.puts "#{typeid}, #{interstatecount},  #{net.test_complexity(
				:IterateCount=>interstatecount.to_i)}"


		else
			output.puts "#{typeid}, #{net.test_tolerance(tolerance_opts)}"
		end
        output.flush
    end

    output.close

else
    puts "Invalid experiment!"
    exit
end

