#!/usr/bin/env ruby

def graph filename, param = ""

	pipe = IO.popen("R --slave", "w+")

	pipe.puts "pdf('#{filename}', family=\"Times\", #{param})"
	pipe.puts yield

	pipe.close_write
	puts pipe.readline while not pipe.eof?
	puts "Wrote #{filename}"

end

def process filename
	file = File.read(filename)
	file.gsub!(/Threshold\d+/, "Threshold")
	file.gsub!(/And/, "AndOrXor")
	file.gsub!(/f0.0-/, "")
	return file
end

def boxplot input, output, label
	graph(output, "width=10") do

		File.open("/tmp/graphinput", "w") do |pipe|
			pipe.write(input)
		end

		%Q(

		par(mar=c(5, 13, 2, 0.5))
		d <-read.csv("/tmp/graphinput");
		names(d)<-c("type", "count");
		boxplot(d$count ~ d$type, horizontal=1,las=1, xlab="#{label}");

		)
	end
end

graph("pdf/complexity.pdf", "width=10") do

	File.open("/tmp/graphinput", "w") do |pipe|
		pipe.write(process("data/intercompl-all"))
	end

	%Q(

		par(mar=c(5, 13, 2, 0.5))
		d <-read.csv("/tmp/graphinput");
		names(d)<-c("type", "intermediary", "complexity");
		boxplot(complexity ~ type, data=d, horizontal=1,las=1, xlab="Complexity");

	)
end

graph("pdf/intercompl.pdf", "height = 10") do

	File.open("/tmp/graphinput", "w") do |pipe|
		pipe.write(process("data/intercompl-all"))
	end

	%Q(
		library(lattice)
		#par(mar=c(5, 13, 2, 0.5))
		d <- read.csv("/tmp/graphinput");
		names(d)<-c("type", "intermediary", "complexity");
		xyplot(intermediary ~ complexity, 
			data=d, 
			groups=type, 
			auto.key=TRUE, 
			xlab="Complexity", 
			ylab="Intermediary state count");

	)
end

graph("pdf/wolfram-1dca.pdf", "height=10") do

	File.open("/tmp/graphinput", "w") do |pipe|
		pipe.write(process("data/wolfram-complexity"))
	end

	%Q(
		d <- read.csv("/tmp/graphinput");
		names(d)<-c("Ruleset", "Complexity");
		boxplot(Complexity ~ Ruleset, data=d, horizontal=1,las=1, yaxt="n");
		axis(2, at=c(25, 50, 75, 100, 125, 150, 175, 200, 225, 250))
		mtext("Rule", side=2, line=3)
		mtext("Complexity", side=1, line=3)

		text(subset(d, Ruleset==68)$Complexity[1], 68, labels="Simple rule",adj=c(0,1))
		points(subset(d, Ruleset==68)$Complexity[1], 68, pch="O", col="red")

		text(subset(d, Ruleset==120)$Complexity[1], 120, labels="Complex rule",adj=c(1,.8))
		points(subset(d, Ruleset==120)$Complexity[1], 120, pch="O", col="red")
	)
end

boxplot(process("data/basincount-all"), "pdf/basincount.pdf", "Basin count")
boxplot(process("data/filtered-basincount-all"), "pdf/filter-basincount.pdf", "Basin count")
boxplot(process("data/intermediary-all"), "pdf/intermediary.pdf", "Intermediary state count")
#boxplot(process("data/robustness-all"), "pdf/robustness.pdf")


