#!/usr/bin/ruby

require "tablefunction.rb"
require "boolnetwork.rb"
require 'yeastnetwork.rb'
require 'typenetwork.rb'
require 'util.rb'
require 'Qt'

#require 'ruby-debug'

$networksize = fromarg(0, 10).to_i
$K = fromarg(1, 3).to_i

$draw_to_terminal = false
$carule = 18

if true
	net = create_random_network($K, $networksize)
else
    #net = TypeNetwork.generate_2d_ca(5)
	#net = YeastNetwork.new("yeast.gv")
	#net = TypeNetwork.load_network("./read.gv")

    #cafunc = lambda{ |states|
        #statepos = states[0] + (states[1] << 1) + (states[2] << 2)
        #if ($carule & (1<<statepos)) == 0
            #0
        #else
            #1
        #end
    #}

    #net = create_1d_ca($networksize, cafunc)
    #$types = [:And, :Or, :Xor]
    #net = TypeNetwork.generate_ca(5, 2)
    #net.show
end

#net.writeGraph("graph.gv")

#for all nodes
#$networksize.times do |x|

	#connect up all inputs
	#$K.times do |inputnumber|
		#node = (rand * $networksize).to_i
		#while net.matrix[x][node] != nil
			#node = (rand * $networksize).to_i
		#end
		#net.matrix[x][node] = inputnumber
	#end
#end

class Ui_MainWindow < Qt::MainWindow

    slots 'on_timerBox_toggled(bool)',
        'on_input_returnPressed()'


    class MatrixWidget < Qt::Widget

        $Boxwidth = 4
        $Boxcountlimit = 120

        def initialize parent
            super(parent)
            @rows = []
        end

        def paintEvent event
            p = Qt::Painter.new(self)
            p.setBrush(Qt::Brush.new(Qt::SolidPattern))

            @rows.each do |row|

                row.each_with_index do |state, i|
                    if state == 1
                        p.drawRect(i*$Boxwidth,0, $Boxwidth, $Boxwidth)
                    end
                end
                p.translate(0, $Boxwidth)
            end

            p.dispose
        end

        def addRow row
            if $draw_to_terminal
                row.each do |state|
                    print " " if state == 0
                    print "X" if state == 1
                end
                puts
                return
            end

            @rows << row
            if @rows.size > $Boxcountlimit
                @rows.delete_at(0)
            end
            update
        end

    end


    def initialize net
        super(nil)
        require 'ui_main.rb'
        setupUi(self)

        @timer = Qt::Timer.new(self)
        @timer.connect(SIGNAL('timeout()')) do
            if @asyncCheck.isChecked
                async_iterate
            elsif @syncCheck.isChecked
                sync_iterate
            end
        end

        @net = net

        @matrixWidget = MatrixWidget.new(@networkBox)
        @networkBox.layout.addWidget(@matrixWidget)

        @functions = [
            ["iterate", "Perform a single iteration", lambda { |args|

            args[0] ||= "async"

            case args[0]
            when "sync"
                sync_iterate
            when "async"
                async_iterate
            end
            }],

            ["reset", "Reset nodes to a random state", Proc.new do
                reset
            end ],

            ["timer", "Toggle timer", Proc.new do
                @timerBox.setChecked(!@timerBox.isChecked)
                on_timerBox_toggled @timerBox.isChecked
            end ],

            ["ls", "List all nodes", Proc.new do
				i = -1
                @net.nodes.map{|n| i += 1; "##{i}: " + n.to_s}.join("\n")
            end ],

            ["description", "Show the description of the currently loaded network", Proc.new do
                @net.description + "\n"
            end ],

            ["find", "Find nodes", Proc.new do |args|
                @net.find(args[0]).map{|n| n.id}.sort.join("\n")
            end ],

            ["carule", "Change carule", Proc.new do |args|
                $carule = (rand * 256).to_i
                "Changed to rule #{$carule}"
            end ],

            ["set", "Set special mode on/off/random. Example: set all on, set all (reset all nodes), set all random",
                Proc.new do |args|
                if args[0] == "all"
                    found = @net.nodes
                else
					found = []
					args[0].split(",").each do |term|
						found << @net.nodes[term.to_i]
					end
                end
                found.each do |node|
                    case args[1]
                    when "on"
                        node.mode = :On
                    when "off"
                        node.mode = :Off
                    when "random"
                        node.mode = :Random
                    else
                        node.mode = nil
                    end
                end
                "OK"
            end ],

            ["save", "Write DOT graph to file (default net.gv)", Proc.new do |args|
				args[0] ||= "net.gv"
                @net.writeGraph(args[0])
				"Written to #{args[0]}"
            end ],

            ["genbn", "Generate a new boolean network with given size (default 20)", Proc.new do |args|
                #@net = create_random_network($K, $networksize)
				args[0] ||= 20
                @net = TypeNetwork.generate_network(args[0].to_i)
				reset
				@net.description
            end ],

            ["genca", "Generate a new cellular automata with given width (default 5)", Proc.new do |args|
                #@net = create_random_network($K, $networksize)
				args[0] ||= 5
				args[1] ||= 2
				args[2] ||= "false"

				width = args[0].to_i
				dim = args[1].to_i
				if args[2] == "true"
					types = [:NThreshold1, 
						:NThreshold2,
						:NThreshold3,
						:NThreshold4,
						:NThreshold5,
						:NThreshold6
					]
				else
					types = [:Threshold1, 
						:Threshold2,
						:Threshold3,
						:Threshold4,
						:Threshold5,
						:Threshold6
					]
				end

                @net = TypeNetwork.generate_ca(width, types, dim)
				reset
				@net.description
            end ],

            ["help", "This thingy", Proc.new do
                @functions.map{|e| e[0] + ": " + e[1]}.sort.join("\n")
            end ],

            ["load", "Load typenetwork from file (default net.gv)", Proc.new do |args|
				args[0] ||= "net.gv"
				@net = TypeNetwork.load_network(args[0])
				@net.description
            end ],

            ["show", "Show the network", Proc.new do |args|
				@net.show
				"OK"
            end ],

            ["change", "Randomize X nodes (default 2)", Proc.new do |args|
				@net.randomize_nodes(args[0])
            end ],

            ["quick", "Toggle drawing to terminal", Proc.new do
                $draw_to_terminal = (not $draw_to_terminal)
                "draw_to_terminal = " + $draw_to_terminal.to_s
            end ],

        ]

        updateNetwork

        @prev_command = "help"

        @input.setFocus

    end




    def handle_command command

        if command == "repeat"
            command = @prev_command
        else
            @prev_command = command
        end

        args = command.split(" ")
        cmd = args.shift

        match = @functions.find_all{ |it|
            it[0].start_with?(cmd)
        }

        if match.size > 1
            return "Ambigious: " + match.map{|e| e[0]}.join(", ")
        elsif match.size == 0
            return "Unknown command"
        else
            begin
                output = match[0][2].call(args)
                output ||= "OK"
                return "> " + command + "\n" + output
            rescue StandardError => e
                return "Error: " + e.to_s
            end
        end
    end

    def on_input_returnPressed
		cmds = input.text.split(";")
		if cmds.empty?
			cmds << "repeat"
		end

		cmds.each do |partcmd|
			output = handle_command(partcmd.strip)
			console.setPlainText(console.toPlainText + output + "\n")
			if output.start_with?("Ambigious")
				break
			else
				input.clear
			end
		end

		sb = console.verticalScrollBar
		sb.setValue(sb.maximum)
    end


    def async_iterate
        @net.iterate :Asynchronous
        updateNetwork
    end

    def sync_iterate
        @net.iterate :Synchronous
        updateNetwork
    end

    def reset
		@net.set_random_state
        updateNetwork
    end

    def on_timerBox_toggled toggled
        if toggled == true
            @timer.start(@timerPrecision.value)
        else
            @timer.stop
        end
    end

    def updateNetwork
        @matrixWidget.addRow(@net.statemap)
    end

end

app = Qt::Application.new(ARGV)

mainw = Ui_MainWindow.new(net)
mainw.show
app.exec
