require 'boolfunction.rb'
require 'boolnetwork.rb'
require 'rubygems'
require 'graphviz'

#network size is 1x1
$distance_threshold = 0.2

#hack to access nodes,edges
class GraphViz
	attr_reader :hoNodes, :loEdges
	class Edge
		attr_reader :xNodeOne, :xNodeTwo
	end
	class Node

	end
end

class TypeNetwork < BooleanNetwork

	def initialize
		super
	end

	def randomize_nodes count = nil
		count ||= 2
		count.times do
			n = nodes.random_item
			#puts "Randomizing #{n.id}"
			n.randomize_inputs self
			n.randomize_func
		end
		"Randomized #{count} nodes"
	end

	def self.load_network filename
		newnet = TypeNetwork.new

		g = GraphViz.parse(filename, {})
		raise "Error: #{filename} is invalid graphviz file" if g.nil?

		ourNodes = {}

		g.hoNodes.each do |name, val|

			#uff så stygt. burde jeg virkelig bruke ruby-graphviz?
			begin
				comment = val["comment"].to_s
				comment.gsub!("\"","")
			rescue TypeError
				comment = "Xor"
			end

			node = TypeFunction.new(comment.to_sym, 3)
			node.name = name

			newnet.addNode(node)
			ourNodes[name] = node
		end

		g.loEdges.each do |edge|
			n0 = ourNodes[edge.xNodeOne]
			n1 = ourNodes[edge.xNodeTwo]

			n0.addInput(n1)
		end

		newnet.description = "TypeNetwork loaded from #{filename}"
		return newnet
	end

	def self.generate_network nodecount, options = nil

		newnet = TypeNetwork.new
        options ||= {}

        k = options[:K] || 3

        $types = options[:Types] || [:Threshold1, :Threshold2, :Threshold3]

        nodecount.times do
            newnet.addNode(TypeFunction.new(nil, k))
        end

        newnet.nodes.each do |node|
			node.randomize_inputs newnet
        end
		
		newnet.description = "TypeNetwork generated with nodecount #{nodecount}"
		return newnet
	end

    def self.generate_ca width, types, dim = 2
		newnet = TypeNetwork.new

        hash = {}

        coord = Array.new(dim, 0)
        (width**dim).times do
            node = newnet.addNode(TypeFunction.new(types.random_item, 2*dim + 1))

            hash[coord.clone] = node
            #puts coord.join(" ")

            place = coord.size - 1
            loop do
                coord[place] +=1
                if coord[place] == width
                    coord[place] = 0
                    place -= 1
                    next
                else
                    break
                end
            end
        end


        def self.wrap(val, max)
            if val < 0
                max - 1
            elsif val >= max
                0
            else
                val
            end
        end

        hash.keys.each do |key|
            inputindex = 1
            node = hash[key]
            node.inputs[0] = node

            dim.times do |dimension|
                [key[dimension] - 1, key[dimension] + 1].each do |val|
                    coord = key.clone
                    coord[dimension] = self.wrap(val, width)
                    #puts coord.join(" ")
                    node.inputs[inputindex] = hash[coord]
                    inputindex += 1
                end
            end
        end

        #loop do
            #newnet.nodes[i].inputs[0] = 
            #newnet.nodes[makeindex(width, x + 1, y)]

            #newnet.nodes[i].inputs[1] = 
            #newnet.nodes[makeindex(width, x, y + 1)]

            #newnet.nodes[i].inputs[2] = 
            #newnet.nodes[makeindex(width, x - 1, y)]

            #newnet.nodes[i].inputs[3] = 
            #newnet.nodes[makeindex(width, x, y - 1)]
        #end

        return newnet
    end

    def self.generate_2d_ca width
		newnet = TypeNetwork.new

        (width*width).times do
            newnet.addNode(TypeFunction.new(nil, 4))
        end

        def self.makeindex(width, x, y)
            x = width if x < 0
            y = width if y < 0

            x = 0 if x >= width
            y = 0 if y >= width

            return y*width + x
        end

        (width*width).times do |i|
            y = i / width
            x = i % width

            newnet.nodes[i].inputs[0] = 
                newnet.nodes[makeindex(width, x + 1, y)]

            newnet.nodes[i].inputs[1] = 
                newnet.nodes[makeindex(width, x, y + 1)]

            newnet.nodes[i].inputs[2] = 
                newnet.nodes[makeindex(width, x - 1, y)]

            newnet.nodes[i].inputs[3] = 
                newnet.nodes[makeindex(width, x, y - 1)]

        end
        return newnet

    end
	

end

$types = nil
#@@types = [:Or, :And, :Xor]

class TypeFunction < BooleanFunction

    attr_reader :x, :y, :type

    def initialize type, numinputs
        super(numinputs, 0)

		if type.nil?
			randomize_func
		else
			set_func(type)
		end


        @name = type.to_s + id

        @x = rand
        @y = rand

    end

	def initialize_copy original
		super
		@x = original.x
		@y = original.y
		@name = original.name
		@type = original.type
	end

	def to_s
		super + " (#{@type.to_s})"
	end

    def write_node graph
		super
		@graphnode.comment = @type.to_s
    end

    def distance_to node
        x = @x - node.x
        y = @y - node.y
        Math.sqrt(x*x + y*y)
    end

	def set_func type
		@type = type

		def make_threshold_func threshold, negative = false
			tryval = (negative)?0:1
            return lambda { |states|
				sum = 0
                states.each do |s|
					sum += 1 if s == tryval
                end
                if sum >= threshold
					1
                else
					0
                end
            }
		end

        case type
        when :Or
            @function = lambda { |states|
                r = 0
                states.each do |s|
                    break if (r |= s) != 0
                end
                if r == 0
                    0
                else
                    1
                end
            }
        when :And
            @function = lambda { |states|
                r = 0
                states.each do |s|
                    break if (r &= s) != 0
                end
                if r == 0
                    0
                else
                    1
                end
            }
        when :Xor
            @function = lambda { |states|
                r = 0
                states.each do |s|
                    r ^= s
                end
                if r == 0
                    0
                else
                    1
                end
            }
		when :Threshold1
            @function = make_threshold_func(1, false)
		when :Threshold2
            @function = make_threshold_func(2, false)
		when :Threshold3
            @function = make_threshold_func(3, false)
		when :Threshold4
            @function = make_threshold_func(4, false)
		when :Threshold5
            @function = make_threshold_func(5, false)
		when :Threshold6
            @function = make_threshold_func(6, false)
		when :Threshold7
            @function = make_threshold_func(7, false)
		when :NThreshold1
            @function = make_threshold_func(1, true)
		when :NThreshold2
            @function = make_threshold_func(2, true)
		when :NThreshold3
            @function = make_threshold_func(3, true)
		when :NThreshold4
            @function = make_threshold_func(4, true)
		when :NThreshold5
            @function = make_threshold_func(5, true)
		when :NThreshold6
            @function = make_threshold_func(6, true)
		when :NThreshold7
            @function = make_threshold_func(7, true)
		else
			raise "Error, invalid type #{type}"
        end
	end

	def randomize_func
		type = $types.random_item
		set_func(type)
	end

	def randomize_inputs net
		raise "Invalid Argument" if not net.is_a?(BooleanNetwork)

		inputs.size.times do |input_index|

			threshold = $distance_threshold
			begin
				close_free_node = 

					net.nodes.find_all{|anode| 
					#har koblet til denne allerede
					not inputs.include?(anode) and

					#den er allerede koblet til meg
					not anode.inputs.include?(self) and

					#den ER meg
					not anode == self and

                    #under threshold
					distance_to(anode) < threshold

				}.random_item

				threshold += 0.1
			end while close_free_node.nil? and threshold <= 1.0

			raise "Could not generate valid network." if close_free_node.nil?

			inputs[input_index] = close_free_node

		end
	end

end
